//==========================================
// Matt Pietrek
// Microsoft Systems Journal, April 1998
// Program: LibDump
// FILE: LibSymbolInfo.CPP
//==========================================

// LibSymbolInfo.cpp: implementation of the CLibSymbolInfo class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "LibDump.h"
#include "LibSymbolInfo.h"
#include "MemoryMappedFile.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define MakePtr( cast, ptr, addValue ) (cast)( (DWORD)(ptr) + (DWORD)(addValue))

/////////////////////////////////////////////////////////////////////////////
// CLibSymbolInfo

CLibSymbolInfo::CLibSymbolInfo() :
	m_bDump(FALSE),
	m_bGrep(FALSE),
	m_nOccurences(0),
	m_bMatchCase(FALSE),
	m_bMatchWholeWord(FALSE)
{
}

CLibSymbolInfo::~CLibSymbolInfo()
{
}

//=============================================================================
// Function: DumpSymbols
//
// Parameters:
//      LPTSTR lpszLibPathName    - The library file path name
//      CStdioFile* pFile         - Address of the file in which to dump the symbols
//
// Description:
//
// Given a library file path name, the function dumps the symbol info into the file
// pointed to by pFile.
//=============================================================================
BOOL CLibSymbolInfo::DumpSymbols(LPTSTR lpszLibPathName, CStdioFile* pFile)
{
	if(pFile == NULL)
	{
		ASSERT(pFile != NULL);
		return FALSE;
	}

	if(lpszLibPathName == NULL || pFile->m_hFile == NULL)
	{
		ASSERT(lpszLibPathName != NULL);
		ASSERT(pFile->m_hFile != NULL);
		m_strErrorMsg = "NULL <lpszLibPathName> or Invalid file handle.";
		return FALSE;
	}

	m_bDump = TRUE;
	m_bGrep = FALSE;

	if(!DumpOrGrep(lpszLibPathName, pFile, NULL, NULL))
		return FALSE;

	return TRUE;
}

//=============================================================================
// Function: GrepSymbols
//
// Parameters:
//      LPTSTR lpszLibPathName  - The library file path name
//      CWnd* pWnd				- Address of the output window in which to dump
//								  grepped symbols						
//		CString& sSymbolToGrep	- Symbol to grep
//		const BOOL& bMatchCase	- Specifies whether the search is case sensitive
//
// Description:
//
// Given a library file path name, the function greps the symbol info and dumps
// it to the window pointed to by pWnd
//=============================================================================
BOOL CLibSymbolInfo::GrepSymbols(LPTSTR lpszLibPathName, CWnd* pWnd, const CString& 
	sSymbolToGrep, const BOOL& bMatchCase, const BOOL& bMatchWholeWord)
{
	if(pWnd == NULL)
	{
		ASSERT(pWnd != NULL);
		return FALSE;
	}

	if(lpszLibPathName == NULL || pWnd->m_hWnd == NULL)
	{
		ASSERT(lpszLibPathName != NULL);
		ASSERT(pWnd->m_hWnd != NULL);
		return FALSE;
	}

	m_bDump = FALSE;
	m_bGrep = TRUE;
	m_bMatchCase = bMatchCase;
	m_bMatchWholeWord = bMatchWholeWord;
		
	if(!DumpOrGrep(lpszLibPathName, NULL, pWnd, sSymbolToGrep))
		return FALSE;

	return TRUE;
}

//=============================================================================
// Function: DumpOrGrep
//
// Parameters:
//      As mentioned above.
//
// Description:
//
// Depending on the value specified in <m_bDump>/<m_bGrep> the routine dumps/greps
// the symbo info.
//=============================================================================
BOOL CLibSymbolInfo::DumpOrGrep(LPTSTR lpszLibPathName, CStdioFile* pFile,
	CWnd* pWnd, LPCTSTR lpszSymbolToGrep) 
{
	CString sBuff;
		
	MEMORY_MAPPED_FILE libFile(lpszLibPathName);
	    
    // Ensure that the file mapping worked
    if( FALSE == libFile.IsValid() )
    {
        if(m_bDump)
			m_strErrorMsg.Format(
				_T("Unable to access file <%s>."), lpszLibPathName
				);

		if(m_bGrep)
		{
			sBuff.Format(
				_T("%s: <Unable to access>\r\n"), lpszLibPathName
				);
			UpdateOutputWindow(sBuff, pWnd);
		}

        return FALSE;
    }

	// All COFF libraries start with the string "!<arch>\n".  Verify that this
    // string is at the beginning of the mapped file
    PSTR pArchiveStartString = (PSTR)libFile.GetBase();
    if ( 0 != strncmp( pArchiveStartString, IMAGE_ARCHIVE_START,
                       IMAGE_ARCHIVE_START_SIZE )  )
    {
        if(m_bDump)
			m_strErrorMsg = "Not a valid COFF LIB file.";

		return FALSE;
    }
	
    // Point to the first archive member.  This entry contains the LIB symbols,
    // and immediately follows the archive start string ("!<arch>\n")
    PIMAGE_ARCHIVE_MEMBER_HEADER pMbrHdr;
    pMbrHdr = MakePtr( PIMAGE_ARCHIVE_MEMBER_HEADER, pArchiveStartString,
                        IMAGE_ARCHIVE_START_SIZE );

    // First DWORD after this member header is a symbol count
    PDWORD pcbSymbols = (PDWORD)(pMbrHdr + 1);  // Pointer math!

    // The symbol count is stored in big endian format, so adjust as
    // appropriate for the target architecture
    DWORD cSymbols = ConvertBigEndian( *pcbSymbols );

    // Following the symbol count is an array of offsets to archive members
    // (essentially, embedded .OBJ files)
    PDWORD pMemberOffsets = pcbSymbols + 1;     // Pointer math!

    // Following the array of member offsets is an array of offsets to symbol
    // names.
    PSTR pszSymbolName = MakePtr( PSTR, pMemberOffsets, 4 * cSymbols );

    // Print out a standard header consisting of the .LIB file name and column
    // headers for the output that will follow
    if(m_bDump)
	{
		pFile->WriteString(sBuff);
		sBuff = "Ord/Hnt  Type  Symbol Information\n";
		pFile->WriteString(sBuff);
		sBuff = "=======  ====  ==================\n";
		pFile->WriteString(sBuff);
	}

    //
    // Loop through every symbol in the first archive member
    //      
    for ( unsigned i = 0; i < cSymbols; i++ )
    {
        DWORD offset;

        // The offsets to the archive member that contains the symbol is stored
        // in big endian format, so convert it appropriately.        
        offset = ConvertBigEndian( *pMemberOffsets );

        // Call DisplayLibInfoForSymbol, which figures out what kind of symbol
        // it is.  The "IsRegularLibSymbol" filters out symbols that are
        // internal to the linking process
        if ( IsRegularLibSymbol( pszSymbolName ) )
            DisplayLibInfoForSymbol(pszSymbolName, libFile.GetBase(), offset, pFile,
							pWnd, lpszSymbolToGrep, lpszLibPathName);
                
        // Advanced to the next symbol offset and name.  The MemberOffsets
        // array has fixed length entries, while the symbol names are
        // sequential null-terminated strings
        pMemberOffsets++;
        pszSymbolName += strlen(pszSymbolName) + 1;
    }

	return TRUE;
}

//=============================================================================
// Function: DisplayLibInfoForSymbol
//
// Parameters:
//      PSTR  pszSymbol           - The symbol name
//      PVOID FileBase            - Base address of the memory mapped file
//      DWORD archiveMemberOffset - Offset of archive member containing symbol
//
// Description:
//
// Given a symbol name, and the offset to the archive member that contains it,
// the routine determines what kind of symbol it is (import by name, import by
// ordinal, a GUID), and displays it accordingly.  If the symbol isn't one of
// these type (e.g., it's a regular staticly linked symbol, the routine just
// displays the symbol name.
//=============================================================================
void CLibSymbolInfo::DisplayLibInfoForSymbol( PSTR pszSymbol, PVOID pFileBase, 
	DWORD archiveMemberOffset, CStdioFile* pFile, CWnd* pWnd, LPCTSTR lpszSymbolToGrep,
	LPTSTR lpszLibPathName)
{
    CStringEx sBuff, sSymbolInfo;

	int nIndex = -1;

    // Make a pointer to the archive member containing the symbol.  The
    // archive member is just an embedded .OBJ preceded by a standard archive
    // member header
    //
    PIMAGE_ARCHIVE_MEMBER_HEADER pMbrHdr
        = MakePtr(PIMAGE_ARCHIVE_MEMBER_HEADER, pFileBase, archiveMemberOffset);

    // Get a pointer to the start of the embedded .OBJ file that follows the
    // archive member header.  COFF .OBJ files begin with an IMAGE_FILE_HEADER
    // structure, which is defined in WINNT.H
    PIMAGE_FILE_HEADER pImgFileHdr = (PIMAGE_FILE_HEADER)(pMbrHdr + 1);

    // Following the IMAGE_FILE_HEADER is the IMAGE_SECTION_HEADER array.
    PIMAGE_SECTION_HEADER pSectHdr = (PIMAGE_SECTION_HEADER)(pImgFileHdr + 1);

    PDWORD pIdata5 = 0;     // Pointers to sections that we'll be looking for
    PWORD  pIdata6 = 0;     // in this .OBJ file.  The presence or absence of
    PBYTE  pRdata  = 0;     // certain sections gives hints what kind of symbol
    PVOID  pText   = 0;     // it is.

    BOOL fSaw_rdata = FALSE;
    
    // Loop through all of the IMAGE_SECTION_HEADERS, looking for .idata$5,
    // .idata$6, and .rdata.  As we find them, create a pointer to the raw
    // data for that section.            
    for ( unsigned i = 0; i < pImgFileHdr->NumberOfSections; i++ )
    {
        _bstr_t bstr = pSectHdr->Name;
		if(bstr == _bstr_t("-1"))
		{
			break;
		}

		if ( 0 == strnicmp( (const char *)pSectHdr->Name, ".idata$5",
								IMAGE_SIZEOF_SHORT_NAME) )
		{
			pIdata5 = MakePtr(PDWORD, pImgFileHdr, pSectHdr->PointerToRawData);
		}
		else if ( 0 == strnicmp( (const char *)pSectHdr->Name, ".idata$6",
								 IMAGE_SIZEOF_SHORT_NAME) )
		{
			pIdata6 = MakePtr(PWORD, pImgFileHdr, pSectHdr->PointerToRawData);
		}
		else if ( 0 == strnicmp( (const char *)pSectHdr->Name, ".text",
								 IMAGE_SIZEOF_SHORT_NAME) )
		{
			pText = MakePtr(PVOID, pImgFileHdr, pSectHdr->PointerToRawData);
		}
		else if ( 0 == strnicmp( (const char *)pSectHdr->Name, ".rdata",
								 IMAGE_SIZEOF_SHORT_NAME) )
		{           
			// Hack!  We'll assume that if the .rdata section is the size of
			// a GUID (0x10 bytes) that it's really a GUID.  There's an obvious
			// flaw here, since sometimes multiple GUIDs are contained within
			// a single .rdata section.
			if ( !fSaw_rdata && (sizeof(GUID) == pSectHdr->SizeOfRawData) )
			{
				pRdata = MakePtr(PBYTE, pImgFileHdr,pSectHdr->PointerToRawData);
			}
			else    // If we've already seen an .rdata section, then we're
			{       // probably not looking at a GUID, so set pRdata to NULL
				pRdata = 0;
			}
        
			fSaw_rdata = TRUE;
		}
        
        pSectHdr++;     // Advance to the next section
    }   

    if ( !pIdata5 ) // If no .idata$5 section, it's not an import symbol
    {
        // If an .rdata section is found, but no .text section, assume it's
        // a GUID symbol
        if ( AfxIsValidAddress(pRdata, sizeof(pRdata)) && !pText)
        {
            // Funky code to break apart the 0x10 bytes of a GUID into its
            // component parts for display
            DWORD part1 = *(PDWORD)pRdata;
            WORD  part2 = *(PWORD)(pRdata+4);
            WORD  part3 = *(PWORD)(pRdata+6);
            
            if(m_bDump)
			{
				sBuff.Format(
					_T("         GUID  {%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X} %s\n"),
                    part1, part2, part3,
                    pRdata[8],  pRdata[9],  pRdata[10], pRdata[11], 
                    pRdata[12], pRdata[13], pRdata[14], pRdata[15],
                    pszSymbol );
	
				pFile->WriteString(sBuff);
			}

			if(m_bGrep)
			{
				sSymbolInfo.Format(_T("{%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X} %s"),
								part1, part2, part3,
								pRdata[8],  pRdata[9],  pRdata[10], pRdata[11], 
								pRdata[12], pRdata[13], pRdata[14], pRdata[15],
								pszSymbol );
	
				if(m_bMatchWholeWord)
				{
					int nPos = -1;
					if((nPos = sSymbolInfo.Find(_T(" "))) != -1)
					{
						if(m_bMatchCase)
						{
							if(!sSymbolInfo.Left(nPos).Compare(lpszSymbolToGrep) ||
								!sSymbolInfo.Mid(nPos+1).Compare(lpszSymbolToGrep) ||
								!sSymbolInfo.Compare(lpszSymbolToGrep))
									nIndex = 0;
						}
						else
						{
							if(!sSymbolInfo.Left(nPos).CompareNoCase(lpszSymbolToGrep) ||
								!sSymbolInfo.Mid(nPos+1).CompareNoCase(lpszSymbolToGrep) ||
								!sSymbolInfo.CompareNoCase(lpszSymbolToGrep))
									nIndex = 0;
						}
					}
				}
				else
				{
					if(m_bMatchCase)
						nIndex = sSymbolInfo.Find(lpszSymbolToGrep);
					else
						nIndex = sSymbolInfo.FindNoCase(lpszSymbolToGrep);
				}
			
				if(nIndex != -1)
				{
					sBuff.Format(_T("%s: %s\r\n"), lpszLibPathName, sSymbolInfo);
					UpdateOutputWindow(sBuff, pWnd);
					m_nOccurences++;
				}
			}
        }
        else    // No .rdata section, so just display the symbol name.  This
        {       // will be the case for staticly linked symbols (e.g., the
                // C++ run time library functions
            if(m_bDump)
			{
				sBuff.Format(_T( "               %s\n"), pszSymbol );
				pFile->WriteString(sBuff);
			}

			if(m_bGrep)
			{
				sSymbolInfo.Format(_T("%s"), pszSymbol );

				if(m_bMatchWholeWord)
				{
					if(m_bMatchCase)
					{
						if(!sSymbolInfo.Compare(lpszSymbolToGrep))
							nIndex = 0;
					}
					else
					{
						if(!sSymbolInfo.CompareNoCase(lpszSymbolToGrep))
							nIndex = 0;
					}
				}
				else
				{
					if(m_bMatchCase)
						nIndex = sSymbolInfo.Find(lpszSymbolToGrep);
					else
						nIndex = sSymbolInfo.FindNoCase(lpszSymbolToGrep);
				}
				
				if(nIndex != -1)
				{
					sBuff.Format(_T("%s: %s\r\n"), lpszLibPathName, sSymbolInfo);
					UpdateOutputWindow(sBuff, pWnd);
					m_nOccurences++;
				}
			}
        }
        
        return;
    }

    //
    // If we get here, and .idata$5 section was found, indicating that it's
    // an import symbol
    //
    WORD ordinal = 0xFFFF;
    BOOL fOrdinal = FALSE;
    PSTR pszExportName = 0;

    // If the first DWORD in the .idata$5 section has the high bit set, it's
    // an import by ordinal symbol, and the low WORD is the ordinal.  If it's
    // not an ordinal import, this DWORD is zero.
    if(AfxIsValidAddress(pIdata5, sizeof(pIdata5)))
	{
		if ( *pIdata5 & IMAGE_ORDINAL_FLAG )
		{
			ordinal = (WORD)(*pIdata5 & ~IMAGE_ORDINAL_FLAG);
			fOrdinal = TRUE;
		}
	}

    // If the .idata$6 section is found, the symbol is imported by name
    if ( AfxIsValidAddress(pIdata6, sizeof(pIdata6)) )
    {
        // The first WORD is a "hint" ordinal value.  Immediately following
        // that is the name of the function that will appear in the EXE/DLLs
        // imports table.  Note that this name can (and often is) different
        // from the name of the symbol that the linker actually references.
        // For example, "_TlsGetValue@4" becomes "TlsGetValue"
        pszExportName = (PSTR)(pIdata6 + 1);    // Pointer math

        if ( FALSE == fOrdinal )
            ordinal = *pIdata6;     // Get the hint ordinal
    }

    // Emit the import (or hint) ordinal, what type of import it is
    // (ORDN, NAME), and the name of the referenced symbol.
    if(m_bDump)
	{
		sBuff.Format(_T( "%7u  %s  %s"), ordinal, fOrdinal ? "ORDN" : "NAME", pszSymbol );
		pFile->WriteString(sBuff);

		// If the symbol is imported by name, append the name that will appear
		// in the imports section of the EXE/DLL.
		if ( pszExportName )
		{
			sBuff.Format(_T( " imported as %s"), pszExportName );
			pFile->WriteString(sBuff);
		}

		sBuff.Format(_T("\n"));
		pFile->WriteString(sBuff);
	}

	if(m_bGrep)
	{
		sSymbolInfo.Format(_T( "%s"), pszSymbol );
	
		if(m_bMatchWholeWord)
		{
			if(m_bMatchCase)
			{
				if(!sSymbolInfo.Compare(lpszSymbolToGrep))
					nIndex = 0;
			}
			else
			{
				if(!sSymbolInfo.CompareNoCase(lpszSymbolToGrep))
					nIndex = 0;
			}
		}
		else
		{
			if(m_bMatchCase)
				nIndex = sSymbolInfo.Find(lpszSymbolToGrep);
			else
				nIndex = sSymbolInfo.FindNoCase(lpszSymbolToGrep);
		}
			
		if(nIndex != -1)
		{
			sBuff.Format(_T("%s: %s\r\n"), lpszLibPathName, sSymbolInfo);
			UpdateOutputWindow(sBuff, pWnd);
			m_nOccurences++;
		}
	}
}

//=============================================================================
// Filters out symbols that are internal to the linking process, and which
// the programmer never explicitly sees.
//=============================================================================
BOOL CLibSymbolInfo::IsRegularLibSymbol( PSTR pszSymbolName )
{
    if ( 0 == strncmp( pszSymbolName, "__IMPORT_DESCRIPTOR_", 20 ) )
        return FALSE;

    if ( 0 == strncmp( pszSymbolName, "__NULL_IMPORT_DESCRIPTOR", 24 ) )
        return FALSE;
        
    if ( strstr( pszSymbolName, "_NULL_THUNK_DATA" ) )
        return FALSE;
        
    return TRUE;
}

//=============================================================================
// Converts from big endian to little endian numbers.
//=============================================================================
DWORD CLibSymbolInfo::ConvertBigEndian(DWORD bigEndian)
{
    DWORD temp = 0;

    temp |= bigEndian >> 24;
    temp |= ((bigEndian & 0x00FF0000) >> 8);
    temp |= ((bigEndian & 0x0000FF00) << 8);
    temp |= ((bigEndian & 0x000000FF) << 24);

    return temp;
}

CString CLibSymbolInfo::GetLastError() const
{
	return m_strErrorMsg;
}

//=============================================================================
// Function: UpdateOutputWindow
//
// Parameters:
//      const CString& sBuff    - Buffer containing the grepped symbol info
//      CWnd* pWnd				- Address of the window in which to dump the symbols
//
// Description:
//
// Given a symbol info and a pointer to a window, the routine dumps the grepped
// symbol info into the window pointed to by pWnd
//=============================================================================
void CLibSymbolInfo::UpdateOutputWindow(const CString& sBuff, CWnd* pWnd)
{
	CRichEditCtrl* pRichEdit = reinterpret_cast<CRichEditCtrl*>(pWnd);
	ASSERT(pRichEdit != NULL);
	pRichEdit->SetSel(-1, -1);
	pRichEdit->ReplaceSel(sBuff);
}

//=============================================================================
// Function: GetOccurences
//
// Parameters:
//      <None>
//
// Description:
//
// Returns the total number of matching symbols grepped
//=============================================================================
int CLibSymbolInfo::GetOccurences()	const
{
	return m_nOccurences;
}